package ru.dbotthepony.mc.otm.worldgen

import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import ru.dbotthepony.kommons.util.KOptional

/**
 * This is a "map stack" of sorts, and when you want to push some data to it
 * you need to push new map to stack (by calling [push]) and then write new data to
 * the map returned by latter.
 */
class PlacementVariableMap {
	private val parent: PlacementVariableMap?
	private val variables = Object2ObjectAVLTreeMap<PlacementVariable<*>, KOptional<*>>()

	constructor() {
		parent = null
	}

	private constructor(parent: PlacementVariableMap) {
		this.parent = parent
	}

	private fun <T> recursiveGet(index: PlacementVariable<T>): KOptional<T>? {
		return variables[index] as KOptional<T>? ?: parent?.recursiveGet(index)
	}

	operator fun <T> get(index: PlacementVariable<T>): KOptional<T> {
		return recursiveGet(index) ?: KOptional()
	}

	operator fun <T> set(index: PlacementVariable<T>, value: T): PlacementVariableMap {
		variables[index] = KOptional(value)
		return this
	}

	fun <T> remove(index: PlacementVariable<T>): KOptional<T> {
		val old = variables.put(index, KOptional<T>())
		return old as KOptional<T>? ?: KOptional()
	}

	fun remove(index: Collection<PlacementVariable<*>>) {
		index.forEach { remove(it) }
	}

	fun push(): PlacementVariableMap {
		return PlacementVariableMap(this)
	}

	override fun equals(other: Any?): Boolean {
		return other === this || other is PlacementVariableMap && variables == other.variables && parent == other.parent
	}

	override fun hashCode(): Int {
		return parent.hashCode() * 31 + variables.hashCode() + 5
	}
}
