package ru.dbotthepony.mc.otm.worldgen.feature

import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import net.minecraft.core.BlockPos
import net.minecraft.core.Holder
import net.minecraft.resources.RegistryFileCodec
import net.minecraft.world.level.levelgen.feature.ConfiguredFeature
import net.neoforged.bus.api.IEventBus
import net.neoforged.neoforge.registries.DataPackRegistryEvent
import ru.dbotthepony.mc.otm.registry.MBuiltInRegistries
import ru.dbotthepony.mc.otm.registry.MDeferredRegister
import ru.dbotthepony.mc.otm.registry.MRegistries
import ru.dbotthepony.mc.otm.worldgen.EnhancedPlacementContext
import ru.dbotthepony.mc.otm.worldgen.PlacementPos
import ru.dbotthepony.mc.otm.worldgen.placement.EnhancedPlacement

abstract class EnhancedFeature<FC>(codec: Codec<FC>) {
	abstract fun place(context: EnhancedPlacementContext, config: FC, positions: Set<PlacementPos>, allPositions: Set<PlacementPos>): Boolean

	data class Configured<F : EnhancedFeature<FC>, FC>(val feature: F, val config: FC) {
		fun place(context: EnhancedPlacementContext, positions: Set<PlacementPos>, allPositions: Set<PlacementPos>): Boolean {
			return feature.place(context, config, positions, allPositions)
		}

		fun placement(): ConfiguredWrapper {
			return ConfiguredWrapper(Holder.direct(this))
		}
	}

	data class ConfiguredWrapper(val configured: Holder<Configured<*, *>>) : EnhancedPlacement {
		override fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos> {
			context.place(positions, configured.value())
			return positions
		}

		override val type: EnhancedPlacement.Type<*>
			get() = Companion

		companion object : EnhancedPlacement.Type<ConfiguredWrapper> {
			override val codec: MapCodec<ConfiguredWrapper> by lazy {
				CODEC.xmap(::ConfiguredWrapper, ConfiguredWrapper::configured).fieldOf("feature")
			}
		}
	}

	val codec: MapCodec<Configured<*, FC>> = codec.fieldOf("config").xmap({ Configured(this, it) }, { it.config })
	fun configure(config: FC) = Configured(this, config).placement()

	object Wrapper : EnhancedFeature<Holder<ConfiguredFeature<*, *>>>(ConfiguredFeature.CODEC) {
		override fun place(
			context: EnhancedPlacementContext,
			config: Holder<ConfiguredFeature<*, *>>,
			positions: Set<PlacementPos>,
			allPositions: Set<PlacementPos>
		): Boolean {
			var any = false
			positions.forEach { any = config.value().place(context.level, context.generator, context.random, it.pos) || any }
			return any
		}

		fun configure(config: ConfiguredFeature<*, *>) = configure(Holder.direct(config))
	}

	companion object {
		private val registrar = MDeferredRegister(MRegistries.FEATURE)

		val DIRECT_CODEC: Codec<Configured<*, *>> by lazy {
			MBuiltInRegistries.FEATURE.byNameCodec().dispatch({ it.feature }, { it.codec })
		}

		val CODEC: Codec<Holder<Configured<*, *>>> by lazy {
			RegistryFileCodec.create(MRegistries.CONFIGURED_FEATURE, DIRECT_CODEC)
		}

		init {
			registrar.register("wrapper") { Wrapper }
		}

		private fun registerRegistry(event: DataPackRegistryEvent.NewRegistry) {
			event.dataPackRegistry(MRegistries.CONFIGURED_FEATURE, DIRECT_CODEC)
		}

		internal fun register(bus: IEventBus) {
			registrar.register(bus)
			bus.addListener(::registerRegistry)
		}
	}
}
