package ru.dbotthepony.mc.otm.worldgen.placement

import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import net.neoforged.bus.api.IEventBus
import ru.dbotthepony.mc.otm.registry.MBuiltInRegistries
import ru.dbotthepony.mc.otm.registry.MDeferredRegister
import ru.dbotthepony.mc.otm.registry.MRegistries
import ru.dbotthepony.mc.otm.worldgen.EnhancedPlacementContext
import ru.dbotthepony.mc.otm.worldgen.PlacementPos
import java.util.stream.Collectors

interface EnhancedPlacement {
	interface Type<T : EnhancedPlacement> {
		val codec: MapCodec<T>
	}

	fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos>
	val type: Type<*>

	class Wrapper(val parent: PlacementModifier) : EnhancedPlacement {
		override fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos> {
			return positions.stream()
				.flatMap { (pos, vars) -> parent.getPositions(context.vanillaContext, context.random, pos).map { PlacementPos(it, vars) } }
				.collect(Collectors.toCollection(::ArrayList))
		}

		override val type: Type<*>
			get() = Companion

		companion object : Type<Wrapper> {
			override val codec: MapCodec<Wrapper> = PlacementModifier.CODEC.xmap(::Wrapper, Wrapper::parent).fieldOf("parent")
		}
	}

	object Passthrough : EnhancedPlacement, Type<Passthrough> {
		override fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos> {
			return positions
		}

		override val type: Type<*>
			get() = this

		override val codec: MapCodec<Passthrough> = MapCodec.unit(this)
	}

	companion object {
		private val registrar = MDeferredRegister(MRegistries.PLACEMENT_MODIFIER)

		init {
			registrar.register("wrapper") { Wrapper.Companion }
			registrar.register("passthrough") { Passthrough }
		}

		val CODEC: Codec<EnhancedPlacement> by lazy {
			MBuiltInRegistries.PLACEMENT_MODIFIER.byNameCodec().dispatch({ it.type }, { it.codec })
		}

		internal fun register(bus: IEventBus) {
			registrar.register(bus)
		}
	}
}
