package ru.dbotthepony.mc.otm.worldgen.placement

import com.google.common.collect.ImmutableList
import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.util.StringRepresentable
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import ru.dbotthepony.mc.otm.worldgen.EnhancedPlacementContext
import ru.dbotthepony.mc.otm.worldgen.PlacementPos

/**
 * Or "shard" placement, if you will.
 *
 * Basically, allows multiple [PlacementModifier]s to split/branch off from provided point.
 */
class EnhancedSplitPlacement(
	val mode: Mode,
	val children: List<EnhancedPlacement>,
) : EnhancedPlacement {
	constructor(mode: Mode, vararg children: EnhancedPlacement) : this(mode, ImmutableList.copyOf(children))

	enum class Mode : StringRepresentable {
		COMBINE,
		FORK;

		private val lname = name.lowercase()

		override fun getSerializedName(): String {
			return lname
		}

		companion object {
			val CODEC: Codec<Mode> = StringRepresentable.fromEnum(::values)
		}
	}

	override fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos> {
		if (mode == Mode.COMBINE) {
			val result = ArrayList<PlacementPos>()
			children.forEach { result.addAll(it.evaluate(context.push(), positions)) }
			return result
		} else {
			children.forEach { it.evaluate(context.push(), positions) }
			return positions
		}
	}

	override val type: EnhancedPlacement.Type<*>
		get() = Companion

	companion object : EnhancedPlacement.Type<EnhancedSplitPlacement> {
		override val codec: MapCodec<EnhancedSplitPlacement> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					Mode.CODEC.fieldOf("mode").forGetter(EnhancedSplitPlacement::mode),
					Codec.list(EnhancedPlacement.CODEC).fieldOf("children").forGetter(EnhancedSplitPlacement::children),
				).apply(it, ::EnhancedSplitPlacement)
			}
		}
	}
}
